# coding: utf-8
"""Configuration module."""
from __future__ import annotations

from typing import Any, Tuple

from pydantic import BaseSettings


class Settings(BaseSettings):
    """Application settings."""

    # CORS configuration
    cors_origins: Tuple[str] = ("*",)
    cors_methods: Tuple[str] = ("*",)
    cors_headers: Tuple[str] = ("*",)
    cors_credentials: bool = True

    # Redis database configuration
    redis_host: str
    redis_user: str
    redis_password: str
    redis_db: int = 0
    redis_max_connections: int = 10
    redis_scheme: str = "redis"

    class Config:
        """Extra configuration."""

        env_file = ".env"

    def __eq__(self, other: Any) -> bool:
        """
        Equality.

        :param other: the other object
        :return: `True` if all attributes are equal, `False` otherwise.
        """
        if not isinstance(other, type(self)):
            return False
        return all(
            getattr(other, name) == getattr(self, name) for name in self.__dict__
        )

    def __hash__(self) -> int:
        """Get this instance's hash code."""
        return hash(tuple(val for val in self.__dict__.values()))



# coding: utf-8
"""
This module contains the building blocks of Agent Communication.

The first block is the agent identity, called :class:`AID`. It is unique to each agent.
Then comes :class:`ACLMessage` and templates.
"""
from __future__ import annotations
from dataclasses import dataclass
import datetime
import enum
import abc
import functools
import operator
import copy

from typing import (
    Dict,
    List,
    Optional,
    Sequence,
    MutableMapping,
    Union,
    Any,
    Type,
)

import piaf.exceptions

__all__ = [
    "AID",
    "Performative",
    "ACLMessage",
    "MessageTemplate",
    "MT_ALL",
    "MT_AND",
    "MT_OR",
    "MT_NOT",
    "MT_PERFORMATIVE",
    "MT_SENDER",
    "MT_CONVERSATION_ID",
    "MT_ENCODING",
    "MT_IN_REPLY_TO",
    "MT_LANGUAGE",
    "MT_ONTOLOGY",
    "MT_PROTOCOL",
    "MT_REPLY_TO",
    "MT_REPLY_WITH",
    "Envelope",
    "ReceivedObject",
    "Message",
]


class AID:
    """
    Represents the identification of an agent.

    This class has three fields:

    * ``name``, which is mandatory and has the following form: agent_name@hap_name
    * ``addresses``, which is a list of URLs to reach the agent. This field is optional
      and URL order is relevant as it defines priorities between URLs.
    * ``resolvers``, which is a list of AIDs of name resolvers agents. This field is
      optional and order is relevant as it defines priorities between AIDs.

    According to the the fipa standard, two AIDs are considered equivalent if their
    ``name`` parameters are the same. Full specification is available here:
    http://www.fipa.org/specs/fipa00023/.
    """

    def __init__(
        self, name: str, addresses: Sequence[str] = (), resolvers: Sequence[AID] = ()
    ):
        """
        Create a new AID object.

        The create object will be initialized with the provided values. The
        ``name`` parameter must following this format : agent@hap_name.

        :param name: the agent's name
        :param addresses: an ordered sequence of URLs.
        :param resolvers: an ordered sequence of AIDs.
        :raise ValueError: if the name parameter is not well formed
        """
        if name is None or "@" not in name:
            raise ValueError("fullname must be agent_name@hap_name")

        self._name: str = name
        self.addresses: Sequence[str] = list(addresses)
        self.resolvers: Sequence[AID] = list(resolvers)

    @property
    def short_name(self) -> str:
        """Short agent's name (ie the part before '@' in the name)."""
        return self._name.split("@")[0]

    @property
    def hap_name(self) -> str:
        """Host agent plateform name (ie the part after '@' in the name)."""
        return self._name.split("@")[1]

    @property
    def name(self) -> str:
        """Agent's name."""
        return self._name

    def __eq__(self, value: Any) -> bool:
        """Two AIDs are equal if and only if their name are the same."""
        return (
            value is not None
            and type(value) == type(self)
            and self._name == value._name
        )

    def __hash__(self):
        return hash(self.name)

    def __deepcopy__(self, memo):
        return AID(self.name, tuple(self.addresses), tuple(self.resolvers))

    def __repr__(self) -> str:
        return f"AID({self.name}, {self.addresses}, {self.resolvers})"


class Performative(enum.Enum):
    """
    Performatives available as defined in http://www.fipa.org/specs/fipa00037/.

    These performatives are intended to be used with :class:`ACLMessage`. Here is
    a small summary of what each performative means:

    **ACCEPT_PROPOSAL**
        The action of accepting a previously submitted proposal to perform an action.
    **AGREE**
        The action of agreeing to perform some action, possibly in the future.
    **CANCEL**
        The action of one agent informing another agent that the first agent no longer
        has the intention that the second agent perform some action.
    **CFP**
        The action of calling for proposals to perform a given action.
    **CONFIRM**
        The sender informs the receiver that a given proposition is true, where the
        receiver is known to be uncertain about the proposition.
    **DISCONFIRM**
        The sender informs the receiver that a given proposition is false, where the
        receiver is known to believe, or believe it likely that, the proposition is
        true.
    **FAILURE**
        The action of telling another agent that an action was attempted but the
        attempt failed.
    **INFORM**
        The sender informs the receiver that a given proposition is true.
    **INFORM_IF**
        A macro action for the agent of the action to inform the recipient whether or
        not a proposition is true.
    **INFORM_REF**
        A macro action for sender to inform the receiver the object which corresponds
        to a descriptor, for example, a name.
    **NOT_UNDERSTOOD**
        The sender of the act (for example, i) informs the receiver (for example, j)
        that it perceived that j performed some action, but that i did not understand
        what j just did. A particular common case is that i tells j that i did not
        understand the message that j has just sent to i.
    **PROPAGATE**
        The sender intends that the receiver treat the embedded message as sent
        directly to the receiver, and wants the receiver to identify the agents denoted
        by the given descriptor and send the received propagate message to them.
    **PROPOSE**
        The action of submitting a proposal to perform a certain action, given certain
        preconditions.
    **PROXY**
        The sender wants the receiver to select target agents denoted by a given
        description and to send an embedded message to them.
    **QUERY_IF**
        The action of asking another agent whether or not a given proposition is true.
    **QUERY_REF**
        The action of asking another agent for the object referred to by a referential
        expression.
    **REFUSE**
        The action of refusing to perform a given action, and explaining the reason for
        the refusal.
    **REJECT_PROPOSAL**
        The action of rejecting a proposal to perform some action during a negotiation.
    **REQUEST**
        The sender requests the receiver to perform some action. One important class of
        uses of the request act is to request the receiver to perform another
        communicative act.
    **REQUEST_WHEN**
        The sender wants the receiver to perform some action when some given
        proposition becomes true.
    **REQUEST_WHENEVER**
        The sender wants the receiver to perform some action as soon as some
        proposition becomes true and thereafter each time the proposition becomes true
        again.
    **SUBSCRIBE**
        The act of requesting a persistent intention to notify the sender of the value
        of a reference, and to notify again whenever the object identified by the
        reference changes.
    """

    def _generate_next_value_(name, start, count, last_values):  # noqa
        return name.lower().replace("_", "-")

    ACCEPT_PROPOSAL = enum.auto()
    AGREE = enum.auto()
    CANCEL = enum.auto()
    CFP = enum.auto()
    CONFIRM = enum.auto()
    DISCONFIRM = enum.auto()
    FAILURE = enum.auto()
    INFORM = enum.auto()
    INFORM_IF = enum.auto()
    INFORM_REF = enum.auto()
    NOT_UNDERSTOOD = enum.auto()
    PROPAGATE = enum.auto()
    PROPOSE = enum.auto()
    PROXY = enum.auto()
    QUERY_IF = enum.auto()
    QUERY_REF = enum.auto()
    REFUSE = enum.auto()
    REJECT_PROPOSAL = enum.auto()
    REQUEST = enum.auto()
    REQUEST_WHEN = enum.auto()
    REQUEST_WHENEVER = enum.auto()
    SUBSCRIBE = enum.auto()


class ACLMessage:
    """
    Representation of an ACL Message as described in
    http://www.fipa.org/specs/fipa00061/.

    FIPA-defined fields are exposed as attributes. Custom fields (aka non FIPA-defined)
    are supported. You can access them using :func:`getattr` function. Those fields are
    prefixed with 'X-'.

    For example, if your message as a custom field named 'custom', you can access to it
    like this::

        value = getattr(msg, "X-custom")

    If you want to build a new :class:`ACLMessage`, use the provided builder
    :class:`ACLMessage.Builder`. Only the builder will make checks to ensure your
    message respects the fipa specification.
    """

    class Builder:
        """
        Ease :class:`ACLMessage` creation.

        It also perform some checks when the message is built. When building a message,
        you must:

        * Instantiate a new builder
        * use :meth:`performative` to set the message performative
        * use whatever method to fill the message
        * invoke the :meth:`build` method

        Calling twice the same method will override the previously set value.
        """

        def __init__(self):
            """Create a new :class:`ACLMessage.Builder`."""
            self._params: MutableMapping[str, Any] = {}

        def performative(self, value: Union[str, Performative]) -> ACLMessage.Builder:
            """Set the message performative."""
            self._params["performative"] = value
            return self

        def sender(self, value: AID) -> ACLMessage.Builder:
            """Set the message sender."""
            self._params["sender"] = value
            return self

        def receiver(self, value: Union[AID, Sequence[AID]]) -> ACLMessage.Builder:
            """Set the message receiver."""
            self._params["receiver"] = value
            return self

        def reply_to(self, value: AID) -> ACLMessage.Builder:
            """Set the reply_to field."""
            self._params["reply_to"] = value
            return self

        def content(self, value: Any) -> ACLMessage.Builder:
            """Set the message content."""
            self._params["content"] = value
            return self

        def language(self, value: str) -> ACLMessage.Builder:
            """Set the message language."""
            self._params["language"] = value
            return self

        def encoding(self, value: str) -> ACLMessage.Builder:
            """Set the message encoding."""
            self._params["encoding"] = value
            return self

        def ontology(self, value: str) -> ACLMessage.Builder:
            """Set the message ontology."""
            self._params["ontology"] = value
            return self

        def protocol(self, value: str) -> ACLMessage.Builder:
            """
            Set the message protocol.

            When set, you should also set the conversation-id and the reply-by fields.
            Otherwise the :meth:`ACLMessage.Builder.build` method will fail.
            """
            self._params["protocol"] = value
            return self

        def conversation_id(self, value: str) -> ACLMessage.Builder:
            """Set the message conversation id."""
            self._params["conversation_id"] = value
            return self

        def reply_with(self, value: str) -> ACLMessage.Builder:
            """Set the reply_with field."""
            self._params["reply_with"] = value
            return self

        def in_reply_to(self, value: str) -> ACLMessage.Builder:
            """Set the in_reply_to field."""
            self._params["in_reply_to"] = value
            return self

        def reply_by(self, value: datetime.datetime) -> ACLMessage.Builder:
            """Set the reply_by field."""
            self._params["reply_by"] = value
            return self

        def custom(self, key: str, value: Any) -> ACLMessage.Builder:
            """
            Add a custom field to the message.

            According to FIPA specification, the key should start with "X-".
            At message build, this will be done if your custom field doesn't already
            starts with it.

            :param key: the key to identify the custom field
            :param value: the value stored in the field
            """
            self._params[key] = value
            return self

        def build(self) -> ACLMessage:
            """
            Build the message.

            Also perform some sanity checks.

            :raise piaf.exceptions.MandatoryFieldValueException: if the performative field is not set
            :raise piaf.exceptions.MandatoryFieldValueException: when ``protocol`` is
                set, ``conversation_id`` and ``reply_by`` should also be set.
            """
            # Perform some checks
            self._check_performative()
            self._check_protocol()
            return ACLMessage(**self._params)

        def _check_performative(self):
            if "performative" not in self._params:
                raise piaf.exceptions.MandatoryFieldValueException("performative")

        def _check_protocol(self):
            if "protocol" in self._params:
                if "conversation_id" not in self._params:
                    raise piaf.exceptions.MandatoryFieldValueException(
                        "conversation_id"
                    )
                if "reply_by" not in self._params:
                    raise piaf.exceptions.MandatoryFieldValueException("reply_by")

    def __init__(
        self,
        performative: Union[str, Performative],
        sender: AID = None,
        receiver: Union[AID, Sequence[AID]] = None,
        reply_to: AID = None,
        content: Any = None,
        language: str = None,
        encoding: str = None,
        ontology: str = None,
        protocol: str = None,
        conversation_id: str = None,
        reply_with: str = None,
        in_reply_to: str = None,
        reply_by: datetime.datetime = None,
        **kwargs: Any,
    ):
        """
        Create a new :class:`ACLMessage`.

        Unless you know what you are doing, use the :class:`ACLMessage.Builder` class
        instead.
        Non fipa fields will be prefixed with 'X-' if not already set.
        """
        self.performative = performative
        self.sender = sender
        self.receiver = receiver
        self.reply_to = reply_to
        self.content = content
        self.language = language
        self.encoding = encoding
        self.ontology = ontology
        self.protocol = protocol
        self.conversation_id = conversation_id
        self.reply_with = reply_with
        self.in_reply_to = in_reply_to
        self.reply_by = reply_by

        # Non FIPA fields : prefix with 'X-' if not already set
        for k, v in kwargs.items():
            if not k.startswith("X-"):
                k = f"X-{k}"
            setattr(self, k, v)

    def __eq__(self, o: object) -> bool:
        """Two :class:`ACLMessage` objects are equals if all fields are equals."""
        if type(o) == type(self):
            for key, val in self.__dict__.items():
                try:
                    if getattr(o, key) != val:
                        return False
                except AttributeError:
                    return False
            return True
        return False

    def __repr__(self) -> str:
        """Textual representation of the ACLMessage for debugging purpose."""
        attrs_and_val = dict(self.__dict__.items())
        text = "ACLMessage("
        text += ",".join(map(lambda k: f"{k}={repr(attrs_and_val[k])}", attrs_and_val))
        return text + ")"


class MessageTemplate(abc.ABC):
    """Common interface for filtering messages."""

    @abc.abstractmethod
    def apply(self, message: ACLMessage) -> bool:
        """
        Apply the template to the provided message.

        :param message: the message to check
        """
        raise NotImplementedError()


class MT_OR(MessageTemplate):
    """Complex template that do a logic "OR" using the provided templates."""

    def __init__(self, a: MessageTemplate, b: MessageTemplate, *args):
        """
        Create a new OR message template with the provided message templates.

        You must provide at least two of them.

        :param a: the first message template
        :param b: the second message template
        :param args: any number of additional message templates
        """
        self._filters = [a, b] + list(args)

    def apply(self, message):
        return functools.reduce(operator.or_, (f.apply(message) for f in self._filters))


class MT_AND(MessageTemplate):
    """Complex template that do a logic "AND" using the provided templates."""

    def __init__(self, a: MessageTemplate, b: MessageTemplate, *args):
        self._filters = [a, b] + list(args)

    def apply(self, message):
        """
        Create a new AND message template with the provided message templates.

        You must provide at least two of them.

        :param a: the first message template
        :param b: the second message template
        :param args: any number of additional message templates
        """
        return functools.reduce(
            operator.and_, (f.apply(message) for f in self._filters)
        )


class MT_NOT(MessageTemplate):
    def __init__(self, template: MessageTemplate) -> None:
        self._template = template

    def apply(self, message: ACLMessage) -> bool:
        return not self._template.apply(message)


def _MT_factory(field_name: str) -> Type[MessageTemplate]:
    """
    Class factory used to create message templates matching simple message fields.

    :param field_name: the message field's name
    """

    def __init__(self, value: str):
        """
        Initializer for the class.

        :param value: the field's value
        """
        self._value = value

    def apply(self, message: ACLMessage) -> bool:
        return getattr(message, field_name) == self._value

    clazz = type(
        f"MT_{field_name.upper()}",
        (MessageTemplate,),
        {"__init__": __init__, "apply": apply},
    )
    return clazz


#: Check the performative field
MT_PERFORMATIVE = _MT_factory("performative")

#: Check the sender field
MT_SENDER = _MT_factory("sender")

#: Check the conversation_id field
MT_CONVERSATION_ID = _MT_factory("conversation_id")

#: Check the encoding field
MT_ENCODING = _MT_factory("encoding")

#: Check the in_reply_to field
MT_IN_REPLY_TO = _MT_factory("in_reply_to")

#: Check the language field
MT_LANGUAGE = _MT_factory("language")

#: Check the ontology field
MT_ONTOLOGY = _MT_factory("ontology")

#: Check the protocol field
MT_PROTOCOL = _MT_factory("protocol")

#: Check the reply_with field
MT_REPLY_WITH = _MT_factory("reply_with")

#: Check the reply_to field
MT_REPLY_TO = _MT_factory("reply_to")


class MT_ALL(MessageTemplate):
    """Special template that matches all messages."""

    def apply(self, message):
        return True


class ReceivedObject:
    """
    Object used in class:`Envelope` object.

    Contains information about the message path, including the ACC who transmitted the message to the receiver.
    See http://www.fipa.org/specs/fipa00067/SC00067F.html
    """

    class Builder:
        """Builder to ease :class:`ReceivedObject` creation."""

        FIPA_PREFIX = "X-"

        def __init__(self) -> None:
            """Create a new empty builder."""
            self._by: Optional[str] = None
            self._date: Optional[datetime.datetime] = None
            self._from: Optional[str] = None
            self._id: Optional[str] = None
            self._via: Optional[str] = None
            self._custom_fields: Dict[str, Any] = {}

        def by(self, by: str) -> ReceivedObject.Builder:
            """
            Set the by field value of the :class:`ReceivedObject` being built.

            Calling this method multiple times will replace previous value.
            """
            self._by = by
            return self

        def date(self, date: datetime.datetime) -> ReceivedObject.Builder:
            """
            Set the date field value of the :class:`ReceivedObject` being built.

            Calling this method multiple times will replace previous value.
            """
            self._date = date
            return self

        def from_(self, from_: str) -> ReceivedObject.Builder:
            """
            Set the from field value of the :class:`ReceivedObject` being built.

            Calling this method multiple times will replace previous value.
            """
            self._from = from_
            return self

        def id(self, id: str) -> ReceivedObject.Builder:
            """
            Set the id field value of the :class:`ReceivedObject` being built.

            Calling this method multiple times will replace previous value.
            """
            self._id = id
            return self

        def via(self, via: str) -> ReceivedObject.Builder:
            """
            Set the via field value of the :class:`ReceivedObject` being built.

            Calling this method multiple times will replace previous value.
            """
            self._via = via
            return self

        def custom(self, fieldname, value) -> ReceivedObject.Builder:
            """
            Add a custom field to the :class:`ReceivedObject` being built.

            The field name must start with FIPA prefix ('X-')
            :raise IllegalArgumentException: If fieldname doesn't start with FIPA prefix
            """
            if not fieldname.startswith(ReceivedObject.Builder.FIPA_PREFIX):
                raise piaf.exceptions.IllegalArgumentException(
                    f"Custom field {fieldname} doesn't start with {ReceivedObject.Builder.FIPA_PREFIX}"
                )
            self._custom_fields[fieldname] = value
            return self

        def build(self) -> ReceivedObject:
            """
            Build the received object.

            :raise MandatoryFieldValueException: if either 'by' or 'date' is not set.
            """
            if self._by is None:
                raise piaf.exceptions.MandatoryFieldValueException("by")
            if self._date is None:
                raise piaf.exceptions.MandatoryFieldValueException("date")
            return ReceivedObject(
                self._by,
                self._date,
                self._from,
                self._id,
                self._via,
                self._custom_fields,
            )

    def __init__(
        self,
        by: str,
        date: datetime.datetime,
        from_: str = None,
        id: str = None,
        via: str = None,
        custom: Dict[str, Any] = None,
    ) -> None:
        """
        Build a new :class:`ReceivedObject`.

        Parameters ``by`` and ``date`` are mandatory. Others can be omitted.

        :param by: URL representing the transport address of the receiving ACC
        :param date: The date when a message was received
        :param from_: The URL representing the transport address of the sending ACC, defaults to None
        :param id: The unique identifier of a message. It is required that uniqueness be garanteed within the scope of
            the sending ACC only, defaults to None
        :param via: The type of the MTP the message was delivered over, defaults to None
        :param custom: Custom fields with their value. Field names must start with FIPA prefix.
        """
        self.by = by
        self.date = date
        self.from_ = from_
        self.id = id
        self.via = via
        if custom is not None:
            self.__dict__.update(custom)

    def __deepcopy__(self, memo):
        custom = {
            name: copy.deepcopy(value, memo)
            for name, value in self.__dict__.items()
            if name.startswith(ReceivedObject.Builder.FIPA_PREFIX)
        }
        return ReceivedObject(
            copy.deepcopy(self.by, memo),
            copy.deepcopy(self.date, memo),
            copy.deepcopy(self.from_, memo),
            copy.deepcopy(self.id, memo),
            copy.deepcopy(self.via, memo),
            custom,
        )


class Envelope:
    """
    Representation of an :class:`ACLMessage` envelope.

    This class supports multiple definition for fields and thus can store the Envelope history.
    Accessing a field directly will always give the last version of it, but you can access to any version
    separately.

    See http://www.fipa.org/specs/fipa00067/SC00067F.html
    """

    class Builder:
        """
        Builder to ease :class:`Envelope` creation.

        There are three types of methods:

         - methods with name matching an Envelope field
         - the :meth:`Builder.with_group_of_fields` method
         - and the :meth:`Builder.build` method

        Each envelope is in fact an history of different fields. To reflect this, the builder
        works that way:

         - the first group of methods defines the history base, ie the value of fields at the creation of the envelope
         - the :meth:`Builder.with_group_of_fields` method adds modifications to fields. Each time the method is called,
           a new "revision" is created.

        In the end, you need to call method :meth:`Builder.build` to build the :class:`Envelope` object.
        """

        TO = "to"
        FROM = "from_"
        ACL_REPRESENTATION = "acl_representation"
        DATE = "date"
        COMMENTS = "comments"
        PLAYLOAD_LENGTH = "playload_length"
        PLAYLOAD_ENCODING = "playload_encoding"
        RECEIVED = "received"
        INTENDED_RECEIVER = "intended_receiver"
        TRANSPORT_BEHAVIOUR = "transport_behaviour"

        # Standard fields
        _FIELDNAMES = [
            TO,
            FROM,
            ACL_REPRESENTATION,
            DATE,
            COMMENTS,
            PLAYLOAD_LENGTH,
            PLAYLOAD_ENCODING,
            RECEIVED,
            INTENDED_RECEIVER,
            TRANSPORT_BEHAVIOUR,
        ]

        # Fipa prefix for custom fields
        _CUSTOM_FIELD_PREFIX = "X-"

        # Mandatory fields in base (revision 0)
        _MANDATORY_FIELDS = ["to", "from_", "acl_representation", "date"]

        def __init__(self) -> None:
            """Create a new :class:`Envelope.Builder` object."""
            self._base: Dict[str, Any] = {}
            self._groups: List[Dict[str, Any]] = [self._base]

        def to(self, to: Sequence[AID]) -> Envelope.Builder:
            """
            Set the ``to`` field of the envelope.

            Each time you call this method, it will replace the old value by the provided one.
            """
            self._base[Envelope.Builder.TO] = to
            return self

        def from_(self, from_: AID) -> Envelope.Builder:
            """
            Set the ``from`` field of the envelope.

            Each time you call this method, it will replace the old value by the provided one.
            """
            self._base[Envelope.Builder.FROM] = from_
            return self

        def acl_representation(self, acl_representation: str) -> Envelope.Builder:
            """
            Set the ``acl-representation`` field of the envelope.

            Each time you call this method, it will replace the old value by the provided one.
            """
            self._base[Envelope.Builder.ACL_REPRESENTATION] = acl_representation
            return self

        def date(self, date: datetime.datetime) -> Envelope.Builder:
            """
            Set the ``date`` field of the envelope.

            Each time you call this method, it will replace the old value by the provided one.
            """
            self._base[Envelope.Builder.DATE] = date
            return self

        def comments(self, comments: str) -> Envelope.Builder:
            """
            Set the ``comments`` field of the envelope.

            Each time you call this method, it will replace the old value by the provided one.
            """
            self._base[Envelope.Builder.COMMENTS] = comments
            return self

        def playload_length(self, playload_length: str) -> Envelope.Builder:
            """
            Set the ``playload-length`` field of the envelope.

            Each time you call this method, it will replace the old value by the provided one.
            """
            self._base[Envelope.Builder.PLAYLOAD_LENGTH] = playload_length
            return self

        def playload_encoding(self, playload_encoding: str) -> Envelope.Builder:
            """
            Set the ``playload-encoding`` field of the envelope.

            Each time you call this method, it will replace the old value by the provided one.
            """
            self._base[Envelope.Builder.PLAYLOAD_ENCODING] = playload_encoding
            return self

        def received(self, received: ReceivedObject) -> Envelope.Builder:
            """
            Set the ``received`` field of the envelope.

            Each time you call this method, it will replace the old value by the provided one.
            """
            self._base[Envelope.Builder.RECEIVED] = received
            return self

        def intended_receiver(
            self, intended_receiver: Sequence[AID]
        ) -> Envelope.Builder:
            """
            Set the ``intended-receiver`` field of the envelope.

            Each time you call this method, it will replace the old value by the provided one.
            """
            self._base[Envelope.Builder.INTENDED_RECEIVER] = intended_receiver
            return self

        def transport_behaviour(self, transport_behaviour) -> Envelope.Builder:
            """
            Set the ``transport-behaviour`` field of the envelope.

            Each time you call this method, it will replace the old value by the provided one.
            """
            self._base[Envelope.Builder.TRANSPORT_BEHAVIOUR] = transport_behaviour
            return self

        def custom(self, fieldname: str, value: Any) -> Envelope.Builder:
            """
            Add a custom field to the envelope.

            If the field was already set, the provided value will replace the old one. Note that according to
            specification, custom fields should start with "X-".
            """
            if not fieldname.startswith(Envelope.Builder._CUSTOM_FIELD_PREFIX):
                raise piaf.exceptions.IllegalArgumentException(
                    f"Custom field {fieldname} doesn't start with {Envelope.Builder._CUSTOM_FIELD_PREFIX}"
                )

            self._base[fieldname] = value
            return self

        def with_group_of_fields(self, group: Dict[str, Any]) -> Envelope.Builder:
            """
            Create a new revision.

            Each time this method is called, a new envelope revision is created. The revision will contain modified
            values + untouched ones.
            For example, if envelope is
            ``Envelope((AID("recv@localhost"),), AID("sender@localhost"), "custom.repr", 01/01/01T01:01)``
            and group is ``{"to": AID("recv2@localhost")}``, then revision 1 (0 is the base) will be
            ``Envelope((AID("recv2@localhost"),), AID("sender@localhost"), "custom.repr", 01/01/01T01:01)``

            Authorized fields are: to, from\\_, acl_representation, date, comments, playload_length, playload_encoding,
            received, intended_receiver, transport_behaviour and custom fields starting by "X-".
            """
            for field in group:
                if field not in Envelope.Builder._FIELDNAMES and not field.startswith(
                    Envelope.Builder._CUSTOM_FIELD_PREFIX
                ):
                    raise piaf.exceptions.IllegalArgumentException(
                        f"Custom field {field} doesn't start with {Envelope.Builder._CUSTOM_FIELD_PREFIX}"
                    )

            self._groups.append(group)
            return self

        def build(self) -> Envelope:
            """Build the envelope."""
            return Envelope(self._groups)

    def __init__(self, groups_of_fields: List[Dict[str, Any]]) -> None:
        """
        Create a new :class:`Envelope` object from the provided groups of fields.

        The first group is the base envelope, ie revision 0. It **MUST** contains the following fields :

        - to: Sequence[AID]
        - from\\_: AID
        - acl_representation: str
        - date: datetime.datetime

        Each additional group defines a new revision.
        Authorized fields are: to, from\\_, acl_representation, date, comments, playload_length, playload_encoding,
        received, intended_receiver, transport_behaviour and custom fields starting by "X-".

        :param group_of_fields: a list of groups
        :raise IndexError: the list must contain at least one group (base)
        :raise MandatoryFieldValueException: if one of the mandatory fields has no value (base group)
        """
        # Check base
        base = groups_of_fields[0]
        for m_field in Envelope.Builder._MANDATORY_FIELDS:
            if m_field not in base:
                raise piaf.exceptions.MandatoryFieldValueException(m_field)

        # Check additional revisions
        self._groups = [base]
        for group in groups_of_fields[1:]:
            self.add_revision(group)

    @property
    def to(self) -> Sequence[AID]:
        """Shortcut access to field to in last revision."""
        return self.get_field_value(Envelope.Builder.TO)

    @property
    def from_(self) -> AID:
        """Shortcut access to field from in last revision."""
        return self.get_field_value(Envelope.Builder.FROM)

    @property
    def acl_representation(self) -> str:
        """Shortcut access to field acl-representation in last revision."""
        return self.get_field_value(Envelope.Builder.ACL_REPRESENTATION)

    @property
    def date(self) -> datetime.datetime:
        """Shortcut access to field date in last revision."""
        return self.get_field_value(Envelope.Builder.DATE)

    @property
    def comments(self) -> str:
        """Shortcut access to field comments in last revision."""
        return self.get_field_value(Envelope.Builder.COMMENTS)

    @property
    def playload_length(self) -> str:
        """Shortcut access to field playload-length in last revision."""
        return self.get_field_value(Envelope.Builder.PLAYLOAD_LENGTH)

    @property
    def playload_encoding(self) -> str:
        """Shortcut access to field playload-encoding in last revision."""
        return self.get_field_value(Envelope.Builder.PLAYLOAD_ENCODING)

    @property
    def received(self) -> ReceivedObject:
        """Shortcut access to field received in last revision."""
        return self.get_field_value(Envelope.Builder.RECEIVED)

    @property
    def intended_receiver(self) -> Sequence[AID]:
        """Shortcut access to field intended-receiver in last revision."""
        return self.get_field_value(Envelope.Builder.INTENDED_RECEIVER)

    @property
    def transport_behaviour(self) -> Any:
        """Shortcut access to field transport-behaviour in last revision."""
        return self.get_field_value(Envelope.Builder.TRANSPORT_BEHAVIOUR)

    @property
    def last_revision_number(self) -> int:
        """Get last revision number."""
        return len(self._groups) - 1

    def get_field_value(self, field: str, revision: int = None) -> Any:
        """
        Get a field's value.

        If revision is None, last revision is assumed. Otherwise, must be an integer greater or equals to 0 (0 meaning
        base envelope).
        If the field doesn't exists, returned value is None (meaning no value associated).
        """
        value = None
        for i, group in enumerate(self._groups):
            if (revision is None or i <= revision) and field in group:
                value = group[field]
        return value

    def get_revision(self, revision: int = None) -> Dict[str, Any]:
        """
        Get a snapshot of fields in the desirated revision.

        If revision is None, last revision is assumed.
        Result is a mapping key -> value.
        """
        result: Dict[str, Any] = {}
        for i, group in enumerate(self._groups):
            if revision is None or i <= revision:
                for field in group:
                    result[field] = group[field]
        return result

    def add_revision(self, group_of_fields: Dict[str, Any]) -> None:
        """
        Create a new revision for this envelope.

        Each time this method is called, a new envelope revision is created. The revision will contain modified
        values + untouched ones.
        For example, if envelope is
        ``Envelope((AID("recv@localhost"),), AID("sender@localhost"), "custom.repr", 01/01/01T01:01)``
        and group is ``{"to": AID("recv2@localhost")}``, then revision 1 (0 is the base) will be
        ``Envelope((AID("recv2@localhost"),), AID("sender@localhost"), "custom.repr", 01/01/01T01:01)``

        Authorized fields are: to, from\\_, acl_representation, date, comments, playload_length, playload_encoding,
        received, intended_receiver, transport_behaviour and custom fields starting by "X-".
        """
        for field in group_of_fields:
            if field not in Envelope.Builder._FIELDNAMES and not field.startswith(
                Envelope.Builder._CUSTOM_FIELD_PREFIX
            ):
                raise piaf.exceptions.IllegalArgumentException(
                    f"Custom field {field} doesn't start with {Envelope.Builder._CUSTOM_FIELD_PREFIX}"
                )

        self._groups.append(group_of_fields)

    def __deepcopy__(self, memo):
        return Envelope(copy.deepcopy(self._groups, memo))


@dataclass(eq=True, frozen=True)
class Message:
    """Data class carrying both the :class:`Envelope` and the :class:`ACLMessage`."""

    envelope: Envelope
    acl_message: ACLMessage

    def __deepcopy__(self, memo):
        envelope = copy.deepcopy(self.envelope, memo)
        acl_message = copy.deepcopy(self.acl_message)

        return Message(envelope, acl_message)



# coding: utf-8
"""
Declare all dependencies in one place.

Dependencies replacement (mainly for test) will be easier.
"""

from functools import lru_cache
from typing import AsyncGenerator

from aioredis import Redis
from fastapi import Depends

from piaf.api.config import Settings
from piaf.api.controllers import PlatformController, RedisController


@lru_cache()
def get_settings() -> Settings:
    """
    Get the application settings.

    :return: application settings
    """
    return Settings()


@lru_cache()
def get_redis_ctrl() -> RedisController:
    """
    Get the application's :class:`RedisController`.

    :return: the redis controller
    """
    ctrl = RedisController(get_settings())
    return ctrl


async def get_redis_session(
    redis_ctrl: RedisController = Depends(get_redis_ctrl),
) -> AsyncGenerator[Redis, None]:
    """
    Yield a session and manage the connection / closing lifecycle.

    :param redis_ctrl: the Redis controller (injected)
    """
    async with redis_ctrl.session() as session:
        yield session


async def get_platform_ctrl(
    settings: Settings = Depends(get_settings),
    redis_session: Redis = Depends(get_redis_session),
) -> PlatformController:
    """
    Get a new :class:`PlatformController` instance.

    :param settings: application settings (injected)
    :param redis_session: a Redis database session (injected)
    """
    return PlatformController(settings, redis_session)



# coding: utf-8
from __future__ import annotations

import asyncio
from time import perf_counter_ns

from piaf.ptf import AgentPlatform
from piaf.agent import Agent

async def main(n: int) -> None:
    ptf = AgentPlatform("localhost")

    async def _launch_agent(name: str) -> None:
        aid = await ptf.agent_manager.create(
            Agent, name
        )
        await ptf.agent_manager.invoke(aid) 


    await ptf.start()
    
    t_start = perf_counter_ns()
    asyncio.gather(*(_launch_agent(f"aget-{i}") for i in range(n)))
    t_stop = perf_counter_ns()
    
    await ptf.stop()

    print(f"Nb of agents: {n}; Elapsed time: {t_stop - t_start}ns")

if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("n", metavar="NB_OF_AGENTS", help="How many agents to launch", type=int)

    parsed = parser.parse_args()
    asyncio.run(main(parsed.n))


# coding: utf-8
"""Utility module."""
# TODO: complete description
from __future__ import annotations
import asyncio
from abc import ABCMeta, abstractmethod
from piaf.behavior import Behavior
from piaf.comm import ACLMessage, MT_OR, MT_PERFORMATIVE, Performative
from typing import Any, TYPE_CHECKING, Dict

if TYPE_CHECKING:
    import piaf.agent


def not_understood_message_from_request(request: ACLMessage, reason: str) -> ACLMessage:
    """
    Build a `NOT_UNDERSTOOD` message from the provided request.

    It will extract the sender and the conversation id of the request. The returned message has performative
    `NOT_UNDERSTOOD` and its content is a list containing two elements:
      - the original request
      - the reason why the request is understood

    :param request: the request message to reply to.
    :param reason: a text message explaining why we reply with `NOT_UNDERSTOOD`
    :return: the built message
    """
    return (
        ACLMessage.Builder()
        .performative(Performative.NOT_UNDERSTOOD)
        .receiver(request.sender)
        .conversation_id(request.conversation_id)
        .content([request, reason])
        .build()
    )


def refuse_message_from_request(request: ACLMessage, reason: str) -> ACLMessage:
    """
    Build a `REFUSE` message from the provided request.

    It will extract the sender and the conversation id of the request. The returned message has performative
    `REFUSE` and its content is a list containing two elements:
      - the original request
      - the reason why the request is refused

    :param request: the request message to reply to.
    :param reason: a text message explaining why we reply with `REFUSE`
    :return: the built message
    """
    return (
        ACLMessage.Builder()
        .performative(Performative.REFUSE)
        .receiver(request.sender)
        .conversation_id(request.conversation_id)
        .content([request.content, reason])
        .build()
    )


def agree_message_from_request(request: ACLMessage) -> ACLMessage:
    """
    Build an `AGREE` message from the provided request.

    It will extract the sender and the conversation id of the request. The returned message has performative
    `AGREE` and its content is set to the content of the request.

    :param request: the request message to reply to.
    :return: the built message
    """
    return (
        ACLMessage.Builder()
        .performative(Performative.AGREE)
        .receiver(request.sender)
        .conversation_id(request.conversation_id)
        .content(request.content)
        .build()
    )


def failure_message_from_request(request: ACLMessage, reason: str) -> ACLMessage:
    """
    Build a `FAILURE` message from the provided request.

    It will extract the sender and the conversation id of the request. The returned message has performative
    `FAILURE` and its content is a list with two items: the first one is set to the content of the request and
    the second one to the provided failure reason.

    :param request: the request message to reply to.
    :param reason: the reason why the operation failed
    :return: the built message
    """
    return (
        ACLMessage.Builder()
        .performative(Performative.FAILURE)
        .receiver(request.sender)
        .conversation_id(request.conversation_id)
        .content([request.content, reason])
        .build()
    )


def inform_message_from_request(request: ACLMessage, content: Any) -> ACLMessage:
    """
    Build an `INFORM` message from the provided request.

    It will extract the sender and the conversation id of the request. The returned message has performative
    `INFORM` and its content is set to the provided one.

    :param request: the request message to reply to.
    :param content: the content of the `INFORM` message
    :return: the built message
    """
    return (
        ACLMessage.Builder()
        .performative(Performative.INFORM)
        .receiver(request.sender)
        .conversation_id(request.conversation_id)
        .content(content)
        .build()
    )


class FIPARequestProtocolBehavior(Behavior, metaclass=ABCMeta):
    """Behavior designed to handle the FIPA request protocol."""

    def __init__(self, agent: piaf.agent.Agent, *args, **kwargs):
        """
        Initialize the :var:`tasks` attribute which stores all running tasks.

        :param agent: the agent that owns the behavior
        """
        super().__init__(agent, *args, **kwargs)
        self.tasks: Dict[str, asyncio.Future] = {}

    def done(self) -> bool:
        """Infinite behavior."""
        return False

    @property
    def msg_template(self) -> piaf.comm.MessageTemplate:
        """
        Template message used to retrieve messages related to the request protocol.

        By default, any message using the :cvar:`Performative.REQUEST` or the :cvar:`Performative.CANCEL` performative
        will match.

        :return: the message template to use
        """
        return MT_OR(
            MT_PERFORMATIVE(Performative.REQUEST), MT_PERFORMATIVE(Performative.CANCEL)
        )

    def check_message(self, msg: ACLMessage) -> bool:
        """
        Check if the given request message is valid.

        Users can override this method to provide their own checks or even send replies. By default it always returns
        `True`.

        :param msg: message to check
        :return: Must return `True` if the message is valid, `False` otherwise.
        """
        return True

    @abstractmethod
    async def on_valid_request(self, msg: ACLMessage) -> None:
        """
        Coroutine called when the request message is valid and can be handled.

        The user should redefine this method. The coroutine will be cancelled if the agent receives a cancel message
        with the same conversation id than the request.

        :param msg: the request ot handle
        :return: coroutine to asynchronously handle it
        """
        raise NotImplementedError()

    async def action(self) -> None:
        """
        Behavior blocks until a valid message (ie validate template :var:`msg_template`) is found.

        Once a message matches, we probe the act and the content to decide what to do.
        """
        # Get next handled message
        msg = await self.agent.receive(self.msg_template)
        acl_msg = msg.acl_message

        # Request -> new conversation
        if acl_msg.performative == Performative.REQUEST:

            # If message content is invalid, stop processing
            if not self.check_message(acl_msg):
                return

            # Otherwise execute the user method
            self.tasks[acl_msg.conversation_id] = asyncio.create_task(
                self.on_valid_request(acl_msg)
            )

        # Cancel -> check if existing conversation
        elif acl_msg.performative == Performative.CANCEL:
            try:
                self.tasks[acl_msg.conversation_id].cancel()
                del self.tasks[acl_msg.conversation_id]
                self.agent.send(inform_message_from_request(acl_msg, acl_msg))
            except KeyError:
                self.agent.send(
                    not_understood_message_from_request(
                        acl_msg, f"Unexpected Act: {acl_msg.performative}"
                    )
                )



# coding: utf-8
"""This module contains a set of helper objects."""
from __future__ import annotations

import asyncio
import importlib
import json
from contextlib import asynccontextmanager
from enum import Enum
from multiprocessing import Process
from typing import TYPE_CHECKING, Any, AsyncGenerator, Dict, List, Tuple, Union, cast

import aioredis
from aioredis import ConnectionPool, Redis

from piaf.agent import Agent
from piaf.api.config import Settings
from piaf.api.exceptions import RedisConnectionNotInitialized, RedisError
from piaf.behavior import Behavior
from piaf.comm import AID, ACLMessage
from piaf.audit import Event, EventRecord, Subscriber, Topic
from piaf.launcher import PlatformLauncher, ServiceDescription
from piaf.ptf import AgentPlatform, AgentPlatformFacade, Extension
from piaf.service import AMSAgentDescription

if TYPE_CHECKING:
    from aioredis.client import PubSub


class APIAgent(Agent):
    """
    A special agent that can communicate with the web API through a Redis connection.

    The agent will first try to connect to the redis database and then will listen to incoming tasks. Tasks execution results are sent back to the web API.
    """

    def __init__(self, aid: AID, platform: AgentPlatformFacade):
        """
        Initialize a new :class:`APIAgent` instance.

        :param aid: the agent's :class:`AID`
        :param platform: a reference to the underlying platform
        """
        super().__init__(aid, platform)
        self.add_behavior(ExecuteTasksBehavior(self))


class ExecuteTasksBehavior(Behavior):
    """
    Pull tasks from the redis channel and execute each one sequentially.

    This behavior requires an open connection to the redis database and a channel listening on the input tasks queue.
    """

    def __init__(self, agent: APIAgent):
        """
        Initialize a new :class:`ExecuteTasksBehavior` instance.

        :param agent: the owner
        """
        super().__init__(agent)

    async def action(self) -> None:
        """
        Wait for incoming tasks and execute each one sequentially.

        Incoming tasks should have the following structure::

            {
                "task_type": "[module.]class",
                "id": "id_of_the_task",
                ...
            }

        Tasks' class are dynamically imported from the specified module or :mod:`piaf.api.tasks` if no module is specified. Results are sent back in json into the response queue. The structure of a response is the following::

            {
                "id": "id-of-the-request",
                "data": whatever is returned by the task's execution or null,
                "error": error message if the execution failed, null otherwise
            }


        """
        async with self.db_session_and_channel() as (client, channel):
            await channel.subscribe(f"channels:{self.agent.aid.hap_name}:from-api")
            async for task in channel.listen():
                json_task: Dict[str, Any] = json.loads(task["data"])
                data = json_task["task_type"]
                split_data = data.rsplit(".", maxsplit=1)

                if len(split_data) == 1:
                    module_name = "piaf.api.impl.tasks"
                    klass_name = split_data[0]
                else:
                    module_name, klass_name = split_data
                module = importlib.import_module(module_name)

                try:
                    klass = getattr(module, klass_name)
                    id_ = json_task["id"]
                    result = await klass.from_json(json_task).execute(self.agent)

                    data = json.dumps(
                        {"id": id_, "data": result, "error": None},
                        default=serialize_piaf_object,
                    )

                    await client.publish(
                        f"channels:{self.agent.aid.hap_name}:to-api",
                        data,
                    )

                except Exception as e:
                    self.agent.logger.exception("Unable to run task.", exc_info=e)
                    await client.publish(
                        f"channels:{self.agent.aid.hap_name}:to-api",
                        json.dumps({"id": id_, "data": None, "error": str(e)}),
                    )

    @asynccontextmanager
    async def db_session_and_channel(
        self,
    ) -> AsyncGenerator[Tuple[Redis, PubSub], None]:
        """
        Work with a Redis client and a channel.

        If the wrapper task is cancelled, this context manager will catch any error, close both the channel and the client and then re-raise it.
        """
        client: None | Redis = None
        channel: None | PubSub = None
        try:
            connection_pool = cast(
                RedisConnectionPoolExtension, self.agent._platform.extensions["redis"]
            )
            client = connection_pool.client
            channel = client.pubsub(ignore_subscribe_messages=True)
            yield client, channel
        finally:
            if channel is not None:
                await channel.close()
            if client is not None:
                await client.close()

    def done(self) -> bool:
        """
        One shot behavior, always return `True`.

        :return: `True`
        """
        return True


class EventsToRedis(Subscriber):
    """
    Subscribe to all events and send them into the redis database.

    This subscriber requires the :class:`RedisConnectionPoolExtension` extension to be loaded in order to send events to the Redis database.
    """

    def __init__(self, platform: AgentPlatform) -> None:
        """
        Create a new :class:`EventToRedis` instance.

        :param platform: the agent platform
        """
        super().__init__()
        self.hap_name = platform.name
        self._platform = platform
        self.redis: None | Redis = None

    async def on_event(self, event_record: EventRecord) -> None:
        """
        Re-publish the event on each queue it is published.

        :param event_record: the record
        """
        if self.redis is None:
            self.redis = cast(
                RedisConnectionPoolExtension, self._platform.extensions["redis"]
            ).client

        json_record = json.dumps(event_record, default=serialize_piaf_object)
        tasks = []
        for channel in event_record.topics:
            tasks.append(
                self.redis.publish(
                    f"channels:{self.hap_name}:events:{channel}", json_record
                )
            )

        results = await asyncio.gather(
            *tasks, return_exceptions=True
        )  # fixme: ignored for now

    async def close(self):
        """
        Close this subscriber.

        It releases the associated Redis connection properly.
        """
        if self.redis is not None:
            await self.redis.close()


#: A type that represents all JSON-compatible types
JSONType = Union[str, int, float, bool, None, List[Any], Dict[str, Any]]


def serialize_piaf_object(
    o: Any,
) -> JSONType:
    """
    Given a `piaf` object, convert it into a JSON-compatible object.

    :param o: the piaf object to serialize
    :return: a JSON-compatible object
    :raise TypeError: the object can't be serialized
    """
    if isinstance(o, AID):
        return {
            "name": o.name,
            "shortname": o.short_name,
            "addresses": o.addresses,
            "resolvers": [serialize_piaf_object(r) for r in o.resolvers],
        }
    if isinstance(o, Enum):
        return o.name
    if isinstance(o, AMSAgentDescription):
        return {
            "aid": serialize_piaf_object(o.name),
            "state": serialize_piaf_object(o.state),
            "owner": o.ownership,
        }
    if isinstance(o, EventRecord):
        return {
            "event": serialize_piaf_object(o.event),
            "topics": [serialize_piaf_object(topic) for topic in o.topics],
            "timestamp": o.timestamp,
        }
    if isinstance(o, Event):
        return {
            "source": o.source,
            "type": o.type,
            "data": json.loads(json.dumps(o.data, default=serialize_piaf_object)),
        }
    if isinstance(o, Topic):
        return str(o)
    if isinstance(o, ACLMessage):
        return {
            k: json.loads(json.dumps(v, default=serialize_piaf_object))
            for k, v in o.__dict__.items()
        }
    return f"Unserializable object: {type(o)}"


class RedisConnectionPoolExtension(Extension):
    """
    An extension providing a pool of Redis connections.

    The user is responsible of closing its session (connection) once the work is finished.
    """

    def __init__(
        self, settings: Settings
    ) -> None:  # TODO: use custom settings, not app ones
        """
        Create a new :class:`RedisConnectionPoolExtension` instance.

        The pool is not bound.

        :param settings: application settings used to setup the pool.
        """
        self._settings = settings
        self._pool: ConnectionPool | None

    async def on_start(self) -> None:
        """
        Initialize the pool and test the connection.

        :raise RedisError: Unable to establish a connection to Redis.
        """
        options: Dict[str, Any] = {
            "decode_responses": True,
            "max_connections": self._settings.redis_max_connections,
        }
        if self._settings.redis_user and self._settings.redis_password:
            options.update(
                {
                    "username": self._settings.redis_user,
                    "password": self._settings.redis_password,
                }
            )
        if self._settings.redis_db:
            options["db"] = self._settings.redis_db

        self._pool = aioredis.ConnectionPool.from_url(
            f"{self._settings.redis_scheme}://{self._settings.redis_host}", **options
        )

        try:
            client: Redis = self.client
            await client.ping()  # Ping to test the connection
            await client.close()
        except Exception as e:
            await client.close()
            if self._pool is not None:
                await self._pool.disconnect()
            raise RedisError() from e

    async def on_stop(self) -> None:
        """Disconnect all connections and close the pool."""
        if self._pool is not None:
            await self._pool.disconnect()

    @property
    def client(self) -> Redis:
        """
        Get a Redis session.

        :raise RedisConnectionNotInitialized: the pool is not bound yet.
        :return: a session
        """
        if self._pool is None:
            raise RedisConnectionNotInitialized()
        return Redis(connection_pool=self._pool)


class AgentPlatformProcess(Process):
    """A customized process that runs a :class:`piaf.ptf.AgentPlatform` on a local asynchronous loop."""

    def __init__(self, name: str, settings: Settings) -> None:
        """
        Create a new instance.

        :param name: name of the platform, which will also be the name of the thread.
        """
        super().__init__(name=name, daemon=True)
        self._settings = settings

    def run(self) -> None:
        """
        Create the asynchronous loop and launch the platform.

        The platform will start with an agent called 'api', which can receive tasks
        from a web interface to execute. For now, logs are streamed in the parent's
        console (if any).
        """
        self.loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self.loop)

        import logging

        logger = logging.getLogger()
        logger.setLevel(logging.INFO)

        handler = logging.StreamHandler()
        handler.setFormatter(
            logging.Formatter("{processName} - {levelname:<8}:{message}", style="{")
        )
        logger.addHandler(handler)

        self.launcher = PlatformLauncher(self.name)
        self.launcher.add_extension(
            "redis", RedisConnectionPoolExtension(self._settings)
        )
        self.launcher.add_service(ServiceDescription("api", APIAgent))

        self.launcher.ptf.evt_manager.subscribe_to(
            EventsToRedis(self.launcher.ptf),
            Topic.from_str(""),
        )

        self.launcher.run()



# coding: utf-8
"""The mod:`piaf.api.routes` modules defines all routes available in the Web API."""
from __future__ import annotations

import asyncio
import json
from typing import TYPE_CHECKING, Any, Dict, List, Union
from uuid import uuid4

import async_timeout as at
from aioredis import Redis
from fastapi import (
    APIRouter,
    Depends,
    Path,
    Query,
    Response,
    WebSocket,
    WebSocketDisconnect,
    status,
)

from piaf.agent import AgentState
from piaf.api.controllers import PlatformController
from piaf.api.dependencies import get_platform_ctrl, get_redis_session
from piaf.api.exceptions import InternalServerError
from piaf.api.models import (
    ACLMessageModel,
    AgentCreationDescriptionModel,
    AgentMemoryModel,
    AgentPlatformModel,
    AgentStateModel,
    AIDModel,
    AMSAgentDescriptionModel,
    ExceptionModel,
)
from piaf.audit import Topic

if TYPE_CHECKING:
    from aioredis.client import PubSub

# V1 router
app_v1 = APIRouter(prefix="/platforms")


@app_v1.post(
    "",
    status_code=status.HTTP_201_CREATED,
    response_description="The platform is created",
    response_model=AgentPlatformModel,
    responses={
        status.HTTP_400_BAD_REQUEST: {
            "model": ExceptionModel,
            "description": "The operation can't be performed.",
            "content": {
                "application/json": {
                    "example": {
                        "detail": "Duplicated platform name my-awesome-platform"
                    }
                }
            },
        }
    },
    tags=["Platform"],
)
async def create_platform(
    ptf: AgentPlatformModel,
    platform_ctrl: PlatformController = Depends(get_platform_ctrl),
):
    """
    Create a new AgentPlatform with the provided name.

    **Body**: a description of the platform to be created.
    """
    try:
        async with at.timeout(5):
            await platform_ctrl.create_platform(ptf)
            return ptf
    except asyncio.TimeoutError as e:
        raise InternalServerError() from e


@app_v1.delete(
    "/{name}",
    response_description="The platform is deleted",
    status_code=status.HTTP_204_NO_CONTENT,
    response_class=Response,
    responses={
        status.HTTP_400_BAD_REQUEST: {
            "model": ExceptionModel,
            "description": "The operation can't be performed.",
        },
    },
    tags=["Platform"],
)
async def delete_platform(
    name: str = Path(description="The platform's name", example="my-awesome-platform"),
    platform_ctrl: PlatformController = Depends(get_platform_ctrl),
):
    """Stop and delete the desired platform."""
    try:
        async with at.timeout(5):
            await platform_ctrl.stop(name)
    except asyncio.TimeoutError:
        await platform_ctrl.kill(name)


@app_v1.get(
    "",
    response_description="Successfully returns the list of active platforms",
    response_model=List[AgentPlatformModel],
    tags=["Platform"],
)
async def get_platforms(
    platform_ctrl: PlatformController = Depends(get_platform_ctrl),
):
    """Get all the running platforms this application is aware of."""
    return await platform_ctrl.get_all_platforms()


@app_v1.post(
    "/{name}/agents",
    status_code=status.HTTP_201_CREATED,
    response_model=AIDModel,
    response_description="The agent is created",
    responses={
        status.HTTP_400_BAD_REQUEST: {
            "model": ExceptionModel,
            "description": "The operation can't be performed.",
        },
    },
    tags=["Agent"],
)
async def create_agent(
    agent: AgentCreationDescriptionModel,
    name: str = Path(description="The platform's name", example="my-awesome-platform"),
    platform_ctrl: PlatformController = Depends(get_platform_ctrl),
):
    """
    Create and invoke an agent into the specified platform.

    **Body** the description of the agent to create
    """
    task = {"task_type": "CreateAgentTask", "id": str(uuid4())}
    task.update(agent.dict())
    return await platform_ctrl.process_task(name, task)


@app_v1.get(
    "/{ptf_name}/agents",
    response_description="Successfully returns the list of agents",
    response_model=List[AMSAgentDescriptionModel],
    responses={
        status.HTTP_400_BAD_REQUEST: {
            "model": ExceptionModel,
            "description": "The operation can't be performed.",
        },
    },
    tags=["Agent"],
)
async def get_agents(
    ptf_name: str = Path(
        description="The platform's name", example="my-awesome-platform"
    ),
    state: Union[None, AgentState] = Query(
        default=None,
        description="Optionally filter results by only keeping agents with the given state.",
        example="ACTIVE",
    ),
    name: Union[None, str] = Query(
        default=None,
        description="Optionally filter results by only keeping agents whose name contains the provided string.",
        example="agent",
    ),
    platform_ctrl: PlatformController = Depends(get_platform_ctrl),
):
    """Retrieve for the given platform all the agents matching the criteria."""
    task = {
        "task_type": "GetAgentsTask",
        "filters": {
            "state": state.name if state is not None else None,
            "name": name if name is not None else "",
        },
        "id": str(uuid4()),
    }
    return await platform_ctrl.process_task(ptf_name, task)


@app_v1.delete(
    "/{ptf_name}/agents/{name}",
    response_description="The agent is deleted",
    status_code=status.HTTP_204_NO_CONTENT,
    response_class=Response,
    responses={
        status.HTTP_400_BAD_REQUEST: {
            "model": ExceptionModel,
            "description": "The operation can't be performed.",
        },
    },
    tags=["Agent"],
)
async def delete_agent(
    ptf_name: str = Path(
        description="The platform's name", example="my-awesome-platform"
    ),
    name: str = Path(
        description="The name of the agent to delete.", example="Custom-1"
    ),
    platform_ctrl: PlatformController = Depends(get_platform_ctrl),
):
    """Delete an agent from the given platform."""
    task = {
        "task_type": "ChangeAgentStateTask",
        "name": name,
        "state": "UNKNOWN",
        "id": str(uuid4()),
    }
    return await platform_ctrl.process_task(ptf_name, task)


@app_v1.get(
    "/{ptf_name}/agents/{name}",
    response_model=AgentMemoryModel,
    response_description="Successfully returns the agent's memory",
    responses={
        status.HTTP_400_BAD_REQUEST: {
            "model": ExceptionModel,
            "description": "The operation can't be performed.",
        },
    },
    tags=["Agent"],
)
async def get_agent_memory(
    ptf_name: str = Path(
        description="The platform's name", example="my-awesome-platform"
    ),
    name: str = Path(description="The name of the agent.", example="Custom-1"),
    platform_ctrl: PlatformController = Depends(get_platform_ctrl),
):
    """Get a snapshot of the current agent's memory."""
    task = {
        "task_type": "RetrieveAgentMemoryTask",
        "aid": {"name": f"{name}@{ptf_name}"},
        "id": str(uuid4()),
    }
    return await platform_ctrl.process_task(ptf_name, task)


@app_v1.post(
    "/{ptf_name}/agents/{name}/messages",
    status_code=status.HTTP_201_CREATED,
    response_description="Successfully sent the message",
    responses={
        status.HTTP_400_BAD_REQUEST: {
            "model": ExceptionModel,
            "description": "The operation can't be performed.",
        },
    },
    tags=["Agent"],
)
async def send_message(
    msg: ACLMessageModel,
    ptf_name: str = Path(
        description="The platform's name", example="my-awesome-platform"
    ),
    name: str = Path(description="The name of the agent.", example="Custom-1"),
    platform_ctrl: PlatformController = Depends(get_platform_ctrl),
):
    """Send a message on the behalf of a specific agent."""
    task = {
        "task_type": "SendMessageTask",
        "id": str(uuid4()),
        "msg": json.loads(msg.json()),
        "sender": f"{name}@{ptf_name}",
    }
    return await platform_ctrl.process_task(ptf_name, task)


@app_v1.put(
    "/{ptf_name}/agents/{name}/state",
    response_description="Successfully updated the agent's state",
    responses={
        status.HTTP_400_BAD_REQUEST: {
            "model": ExceptionModel,
            "description": "The operation can't be performed.",
        },
    },
    tags=["Agent"],
)
async def update_agent_state(
    state: AgentStateModel,
    ptf_name: str = Path(
        description="The platform's name", example="my-awesome-platform"
    ),
    name: str = Path(description="The name of the agent.", example="Custom-1"),
    platform_ctrl: PlatformController = Depends(get_platform_ctrl),
):
    """Replace an agent's state by the provided one."""
    task = {
        "task_type": "ChangeAgentStateTask",
        "name": name,
        "state": state.state.name,
        "id": str(uuid4()),
    }
    return await platform_ctrl.process_task(ptf_name, task)


@app_v1.websocket("/{ptf_name}/ws")
async def topic_listener(
    ptf_name: str,
    websocket: WebSocket,
    redis_session: Redis = Depends(get_redis_session),
) -> None:
    """
    Get a websocket that can listen on the platform's event.

    The websocket supports two methods:

    - subscribe: subscribe to a particular topic
    - unsubscribe: unsubscribe from a particular topic

    Here is the Json object::

        {
            method: "[un]subscribe",
            topic: ".some.topic"
        }

    .. warning:: Contrary to how events are dispatched inside the piaf platform, events are not dispatched to topic's parents. It means that listening on `.platform` won't catch events emitted on `.platform.agents` for example.

    :param ptf_name: the platform's name
    :param websocket: injected by FastAPI
    :param redis_session: a Redis session, injected by FastAPI
    """

    try:
        await websocket.accept()
        pubsub: PubSub = redis_session.pubsub()
        first_sub: bool = True

        while True:
            try:
                data: Dict[str, Any] = await websocket.receive_json()
            except WebSocketDisconnect:
                break

            topic = Topic.from_str(data["topic"])
            if data["method"] == "subscribe":
                await pubsub.subscribe(f"channels:{ptf_name}:events:{topic}")

            if data["method"] == "unsubscribe":
                await pubsub.unsubscribe(f"channels:{ptf_name}:events:{topic}")

            if first_sub:
                task = asyncio.create_task(yield_events(pubsub, websocket))
                first_sub = False
    finally:
        task.cancel()
        await pubsub.close()


async def yield_events(pubsub: PubSub, ws: WebSocket) -> None:
    """
    Asynchronous task that keeps listening on the provided pubsub and yielding downloaded data to the provided websocket.

    :param pubsub: redis publish/subscribe object
    :param ws: websocket
    """
    while True:
        record = await pubsub.get_message(ignore_subscribe_messages=True, timeout=2)
        if record is not None:
            await ws.send_json(json.loads(record["data"]))
        await asyncio.sleep(0)
